#!/usr/bin/env python
"""
This program is used to visualize contour data created by SimVascular. 
"""
import argparse
import logging
import sys
import xml.etree.ElementTree as et
import vtk
import numpy as np
import copy
import csv



class ContourGroup(object):
    """
    The ContourGroup class is used to represent a SimVascular contour group.
    
    Obs.: It is ignoring the timestep tag.
    """
    
    def __init__(self, logger):
        self.logger = logger
        self.attrib = dict()
        self.lofting_parameters = dict()
        self.contours = []
        
    def __repr__(self):
        return '<ContourGroup({0}) id={1}>'.format(self.attrib["path_name"],
                                                   id(self))
    
    def __copy__(self):
        return ContourGroup(self, self.logger)
    
    def __deepcopy__(self, memo):
        newContourGroup = ContourGroup(self.logger)
        newContourGroup.__dict__.update(self.__dict__)
        newContourGroup.attrib = copy.deepcopy(self.attrib, memo)
        newContourGroup.lofting_parameters \
                                = copy.deepcopy(self.lofting_parameters, memo)
        newContourGroup.contours = copy.deepcopy(self.contours, memo)
        
        return newContourGroup
    
    def read(self, file_name):
        """
        Read a ctgr file. 
        """
        
        if file_name == None:
            self.logger.error("No input contour file name given.")
            sys.exit(1)
        else:
            self.logger.info("Input contour file name: {}".format(file_name))
    
        # Remove 'format' tag from xml file.
        f = open(file_name, "r")
        lines = f.readlines()
        new_lines = []
        for line in lines:
          if '<format' not in line:
            new_lines.append(line)
        # Create string from xml file and parse it.
        xml_string = "".join(new_lines)
        
        tree = et.ElementTree(et.fromstring(xml_string))
        contour_group_et = tree.getroot() #the root is contour_group
        
        for attrib, value in contour_group_et.attrib.items():
            self.attrib[attrib] = value
        self.logger.info("Contour group: %s" % self.attrib["path_name"])
        
#        self.attrib["path_id"] = contour_group_et.attrib.get("path_id")
#        self.attrib["reslice_size"] = contour_group_et.get("reslice_size")
#        self.attrib["point_2D_display_size"] = contour_group_et.attrib.get("point_2D_display_size")
#        self.attrib["point_size"] = contour_group_et.attrib.get("point_size")
        
        lofting_parameters_et = contour_group_et.find("timestep/lofting_parameters")
        for attrib, value in lofting_parameters_et.attrib.items():
            self.lofting_parameters[attrib] = value
        
        for contour_et in contour_group_et.iter('contour'):
            contour = Contour()
            #logger.info("   Contour id: %s" % id)
            contour.read(contour_et)
            #_for control_pts in contour_t
            self.contours.append(contour)
        #_for contour_t in contour_group_t
        self.logger.info("   Number of contours %d" % len(self.contours))
    
    def write(self, file_name):
        """
        Write a ctgr file. 
        """
        
        contour_group_et = et.Element("contourgroup")
        for attrib, value in self.attrib.items():
            contour_group_et.set(attrib, str(value))
        
        timestep = et.SubElement(contour_group_et, "timestep")
        timestep.set("id", "0")
        
        lofting_parameters_et = et.SubElement(timestep, 'lofting_parameters')
        for attrib, value in self.lofting_parameters.items():
            lofting_parameters_et.set(attrib, str(value))
        
        for contour in self.contours:
            contour_et = et.SubElement(timestep, 'contour')
            
            
            for attrib, value in contour.attrib.items():
                contour_et.set(attrib, str(value))
            
            # path_point
            path_point_et = et.SubElement(contour_et, 'path_point')
            path_point_et.set("id", str(contour.path_point.id))
            
            pos_et = et.SubElement(path_point_et, 'pos')
            pos_et.set("x", str(contour.path_point.pos[0]))
            pos_et.set("y", str(contour.path_point.pos[1]))
            pos_et.set("z", str(contour.path_point.pos[2]))
            
            tangent_et = et.SubElement(path_point_et, 'tangent')
            tangent_et.set("x", str(contour.path_point.tangent[0]))
            tangent_et.set("y", str(contour.path_point.tangent[1]))
            tangent_et.set("z", str(contour.path_point.tangent[2]))
            
            rotation_et = et.SubElement(path_point_et, 'rotation')
            rotation_et.set("x", str(contour.path_point.rotation[0]))
            rotation_et.set("y", str(contour.path_point.rotation[1]))
            rotation_et.set("z", str(contour.path_point.rotation[2]))
            
            # control_points
            control_points_et = et.SubElement(contour_et, 'control_points')
            for i, point in enumerate(contour.control_points):
                point_et = et.SubElement(control_points_et, 'point')
                point_et.set("id", str(i))
                point_et.set("x", str(point[0]))
                point_et.set("y", str(point[1]))
                point_et.set("z", str(point[2]))
                
            # contour_points
            contour_points_et = et.SubElement(contour_et, 'contour_points')
            for i, point in enumerate(contour.contour_points):
                point_et = et.SubElement(contour_points_et, 'point')
                point_et.set("id", str(i))
                point_et.set("x", str(point[0]))
                point_et.set("y", str(point[1]))
                point_et.set("z", str(point[2]))
        
        indent(contour_group_et)
        with open(file_name, "w") as f:
            f.write(  "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n" \
                    + "<format version=\"1.0\" />\n" \
                    + et.tostring(contour_group_et).decode("utf-8"))

    def edit(self, editDict):
        self.logger.info("Editting using: {}".format(editDict))
        
#        # For debugging
#        perimeter = self.contours[5].getPerimeter()
#        area = self.contours[5].getArea()
#        circularity = self.contours[5].getCircularity()
#        flattening = self.contours[5].getFlattening()
#        meanD = self.contours[5].getMeanDiameter()
#        stdD = self.contours[5].getStdDiameter()
#        print(area, perimeter, circularity, flattening, meanD, stdD)
        
        
        with open(editDict) as f:
            commands = [{k: str(v) for k, v in row.items()}
                        for row in csv.DictReader(f, skipinitialspace=True)]
            
        for command in commands:
            contour_ids = command['contour_id'].split(" ")
            contour_ids = list(map(int, contour_ids)) # list of int
            
            for i in contour_ids:
                if command["operation"] == "changeType":
                    self.contours[i].setType(command["parameters"])
                elif command["operation"] == "scaleRadius":
                    self.contours[i].scaleRadius(command["parameters"])
                elif command["operation"] == "scaleArea":
                    self.contours[i].scaleArea(command["parameters"])
                elif command["operation"] == "setArea":
                    self.contours[i].setArea(command["parameters"])
        
#        perimeter = self.contours[5].getPerimeter()
#        area = self.contours[5].getArea()
#        circularity = self.contours[5].getCircularity()
#        flattening = self.contours[5].getFlattening()
#        meanD = self.contours[5].getMeanDiameter()
#        stdD = self.contours[5].getStdDiameter()
#        print(area, perimeter, circularity, flattening, meanD, stdD)
        
    def savaData(self, csv_file):
        self.logger.info(  "Saving statistics of " \
                         + "geometrical features in: {}".format(csv_file))
        
        csv_columns = ['contour_id',
                       'center_x', 'center_y', 'center_z',
                       'u_x', 'u_y', 'u_z',
                       'v_x', 'v_y', 'v_z',
                       'n_x', 'n_y', 'n_z',
                       'Mean Diameter', 'Std Diameter',
                       'Area', 'Perimeter',
                       'Circularity', 'Flattening']

        
        try:
            with open(csv_file, 'w', newline='') as f:
                writer = csv.DictWriter(f, fieldnames=csv_columns)
                writer.writeheader()
                for contour_id, contour in enumerate(self.contours):
                    contour_data = dict()
                    
                    contour_data['contour_id'] = contour_id
                    
                    cPoint = contour.getCenter()
                    contour_data['center_x'] = cPoint[0]
                    contour_data['center_y'] = cPoint[1]
                    contour_data['center_z'] = cPoint[2]
                    
                    u, v, n = contour.getUVN()
                    contour_data['u_x'] = u[0]
                    contour_data['u_y'] = u[1]
                    contour_data['u_z'] = u[2]
                    contour_data['v_x'] = v[0]
                    contour_data['v_y'] = v[1]
                    contour_data['v_z'] = v[2]
                    contour_data['n_x'] = n[0]
                    contour_data['n_y'] = n[1]
                    contour_data['n_z'] = n[2]
                    
                    contour_data['Mean Diameter'] = contour.getMeanDiameter()
                    contour_data['Std Diameter'] = contour.getStdDiameter()
                    contour_data['Area'] = contour.getArea()
                    contour_data['Perimeter'] = contour.getPerimeter()
                    contour_data['Circularity'] = contour.getCircularity()
                    contour_data['Flattening'] = contour.getFlattening()
                    
                    writer.writerow(contour_data)
        except IOError:
            self.logger.error("I/O error with {}".format(csv_file))

        

class Contour(object):
    """
    The Contour class is used to represent a SimVascular contour.
    """
    def __init__(self):
        self.attrib = dict()
        self.path_point = Path_Point()
        ''' Note
        control_points[0] : center point
        control_points[1] : scale point
        control_points[>1] : when present, they depend on the contour type
        for Ellipse:
            control_points[2] : point at major axis
            control_points[3] : point at minor axis
        for Polynoms and PolySplines:
            control_points[>1] : respectie control points
        '''
        self.control_points = list()
        self.contour_points = list()
    
    def read(self, contour_et):
        
        for attrib, value in contour_et.attrib.items():
            self.attrib[attrib] = value
        
        path_point_et = contour_et.find("./path_point")
        self.path_point.read(path_point_et)
        
        for control_pts in contour_et.iter('control_points'):
            #logger.info("      Control points ")
            for point in control_pts.iter('point'):
                x = point.attrib['x']
                y = point.attrib['y']
                z = point.attrib['z']
                point = np.array([float(x), float(y), float(z)])
                self.control_points.append(point)
        for contour_pts in contour_et.iter('contour_points'):
            #logger.info("      Contour points ")
            for point in contour_pts.iter('point'):
                x = point.attrib['x']
                y = point.attrib['y']
                z = point.attrib['z']
                point = np.array([float(x), float(y), float(z)])
                self.contour_points.append(point)
            #_for point in control_pts
    
    def getUVN(self):
        '''
        Returns the versors of the plane coordinate system.
        
        Returns
        -------
            u : ndarray(3)
                
            v : ndarray(3)
                
            n : ndarray(3)
                normal versor.
        '''
        n = self.path_point.tangent
        n /= np.linalg.norm(n)
        
        u = self.path_point.rotation
        u /= np.linalg.norm(u)
        
        v = np.cross(u, n)
        v /= np.linalg.norm(v)
        
        return (u, v, n)
    
    
    def getCenter(self):
        '''
        Returns the centre of the contour points.
        '''
        #this is different than the contour_points[0]...
        return np.mean(self.contour_points, axis=0)
    
    
    def getPerimeter(self):
        '''
        Returns the perimeter of the contour. The Euclidian distance between 
        countour points is summed up together.
        
        Returns
        -------
            perimeter : float
        '''
        per = 0.0
        for i in range(len(self.contour_points)):
            per += np.linalg.norm(  self.contour_points[i]   \
                                  - self.contour_points[i-1])
        return per
    
    
    def getArea(self):
        '''
        Returns the area of the contour using Green's Theorem.
        
        Returns
        -------
            area : float
        
        References
        ----------
        .. https://en.wikipedia.org/wiki/Green%27s_theorem#Area_calculation
        '''
        
        contour_uvn = toPlaneCoordinates(self.path_point, self.contour_points)
        
        area = 0.0
        for i in range(len(self.contour_points)):
            u = contour_uvn[i-1]
            du = contour_uvn[i] - contour_uvn[i-1]
            
            area += u[0]*du[1] - u[1]*du[0]
        
        area *= 0.5
        
        return np.abs(area)
    
    
    def getCircularity(self):
        '''
        Circularity is the ratio of the integrated area of the cross-section 
        to the equivalent area of a circle of the same perimeter.
        
        Returns
        -------
            circularity : float
        '''
        return 4.0*np.pi*self.getArea()/self.getPerimeter()**2
    
    
    def getMeanDiameter(self):
        '''
        Returns the mean diameter of the contour points.
        '''
        contour_points_nparray = np.array(self.contour_points)
        radii = np.linalg.norm(contour_points_nparray - self.getCenter(),
                               axis=1)
        
        return 2.0*np.mean(radii)
    
    
    def getStdDiameter(self):
        '''
        Returns the standard deviation of the diameter of the contour points.
        '''
        contour_points_nparray = np.array(self.contour_points)
        radii = np.linalg.norm(contour_points_nparray - self.getCenter(),
                               axis=1)
        
        return 2.0*np.std(radii)
    
    
    def getFlattening(self):
        '''
        Flattening is a measure of the compression of a circle along a diameter 
        to form an ellipse. Other terms used are ellipticity or oblateness. The 
        usual notation for flattening is f and its definition in terms of the 
        semi-axes of the resulting ellipse is
    
        .. math::
            f = 1 - \\frac{b}{a}
    
        where a and b are the major and minor axis of the ellipse, respectively. 
        The compression factor is ''b/a'' in each case. For the ellipse, this 
        factor is also the aspect ratio of the ellipse.
    
        Returns
        -------
            flattening : float
            
        References
        ----------
        .. https://en.wikipedia.org/wiki/Flattening
    
        '''
        contour_uvn = toPlaneCoordinates(self.path_point, self.contour_points)
        
        # Best-fitting into an ellipse
        ellipse = Ellipse()
        contour_uvn_nparray = np.array(contour_uvn)
        ellipse.fitEllipse(contour_uvn_nparray[:,0], contour_uvn_nparray[:,1])
        axes = ellipse.getAxisLength()
        
        return  1.0 - axes.min()/axes.max()
    
    
    def setType(self, label):
        '''
        Changes the line type of the closed contour.
        
        Parameters
        ----------
        label : string
            line type
            Available: Circle, Ellipse
        '''
        area = self.getArea()
        center = self.getCenter()
        self.control_points[0] = center
        
        u, v, n = self.getUVN()
        
        contour_uvn = toPlaneCoordinates(self.path_point, self.contour_points)
        
        if label=="Circle":
            self.attrib["type"] = label
            self.attrib["method"] = "Manual"
            self.attrib["closed"] = "true"
            self.attrib["min_control_number"] = 2
            self.attrib["max_control_number"] = 2
            self.attrib["subdivision_type"] = 2
            self.attrib["subdivision_number"] = 36
            self.attrib["subdivision_spacing"] = 0.078125 # change this later!
            
            radius = np.sqrt(area/np.pi)
            
            #reset control_points
            self.control_points = []
            self.control_points.append(center)
            radiusHandler = center + radius*(np.cos(np.pi/4)*u - np.sin(np.pi/4)*v)
            self.control_points.append(radiusHandler)
            
            angles = np.linspace(0.0,
                                 2*np.pi,
                                 self.attrib["subdivision_number"],
                                 endpoint=False)
            self.contour_points = []
            for angle in angles:
                point = center + radius*(np.cos(angle)*u + np.sin(angle)*v)
                self.contour_points.append(point)
        elif label=="Ellipse":
            self.attrib["type"] = label
            self.attrib["method"] = "Manual"
            self.attrib["closed"] = "true"
            self.attrib["min_control_number"] = 4
            self.attrib["max_control_number"] = 4
            self.attrib["subdivision_type"] = 2
            self.attrib["subdivision_number"] = 36
            self.attrib["subdivision_spacing"] = 0.078125 # change this later!
            self.attrib["as_circle"] = "false"
            
            # Best-fitting into an ellipse
            ellipse = Ellipse()
            contour_uvn_nparray = np.array(contour_uvn)
            ellipse.fitEllipse(contour_uvn_nparray[:,0],
                               contour_uvn_nparray[:,1])
            
            axes = ellipse.getAxisLength()
            radius = np.sqrt(axes[0]*axes[1])
            theta = ellipse.getAngleOfRotation()
            
            R = np.array([[np.cos(theta), -np.sin(theta), 0.0],
                          [np.sin(theta),  np.cos(theta), 0.0],
                          [          0.0,            0.0, 1.0]])
            T = np.stack([u,v,n], axis=1)
            T /= np.linalg.det(T)
            
            # direction of the axes of the ellipse in global coordinates
            ulin = np.dot(T, np.dot(R, np.array([1.0, 0.0, 0.0])))
            vlin = np.dot(T, np.dot(R, np.array([0.0, 1.0, 0.0])))
            
            # reset control_points
            scaleHandler = 0.5*radius*(  np.cos(np.pi/4)*ulin 
                                       - np.sin(np.pi/4)*vlin)
            majorAxis = center + axes[0]*(  np.cos(0.0)*ulin 
                                          + np.sin(0.0)*vlin)
            minorAxis = center + axes[1]*(  np.cos(np.pi/2.0)*ulin 
                                          - np.sin(np.pi/2.0)*vlin)
            
            self.control_points = []
            self.control_points.append(center)
            self.control_points.append(scaleHandler)
            self.control_points.append(majorAxis)
            self.control_points.append(minorAxis)
            
            # reset contour_points
            angles = np.linspace(0.0,
                                 2*np.pi,
                                 self.attrib["subdivision_number"],
                                 endpoint=False)
            self.contour_points = []
            for angle in angles:
                point = center + axes[0]*np.cos(angle)*ulin + axes[1]*np.sin(angle)*vlin
                self.contour_points.append(point)
        else:
            print("ERROR: Invalid contour type (\"{}\")!".format(label))
            sys.exit(1)
    
    
    def scaleRadius(self, scale):
        '''
        Linearly scales the contour on the same plane, around the center point.
        
        Parameters
        ----------
        scale : float, int or string
            scale factor.
        '''
        scale = float(scale)
        #skip iteration, if scale = 1.0
        if scale == 1.0:
            return
        else:
            # The first control point is the contour center.
            p0 = self.control_points[0]  #nparray
            # ConTROL points
            for i, control_point in enumerate(self.control_points): 
                self.control_points[i] = scale*(control_point - p0) + p0
            # ConTOUR points
            for i, contour_point in enumerate(self.contour_points):
                self.contour_points[i] = scale*(contour_point - p0) + p0

    def scaleArea(self, scale):
        '''
        Scales the contour area, keeping the same shape around the center point.
        
        Parameters
        ----------
        scale : float, int or string
            scale factor.
        '''
        scale = np.sqrt(float(scale))
        self.scaleRadius(scale)
    
    def setArea(self, area):
        '''
        Scales the contour area into a given cross-sectional area.
        
        Parameters
        ----------
        area : float, int or string
            new cross-sectional area.
        '''
        newArea = float(area)
        oldArea = self.getArea()
        scale = newArea/oldArea
        self.scaleArea(scale)
        
        
class Path(object):
    """
    The Path class is used to represent a SimVascular path.
    """
    
    def __init__(self, path_name):
        self.id = path_name
        self.pos = []
        self.tangent = []
        self.rotation = []

class Path_Point(object):
    """
    The Path_Point class is used to represent a SimVascular path_point.
    """
    
    def __init__(self):
        self.id = None
        self.pos = np.zeros(3)
        self.tangent = np.zeros(3)
        self.rotation = np.zeros(3)
    
    def read(self, path_point_et):
        self.id = path_point_et.attrib["id"]
        
        pos_et = path_point_et.find("./pos")
        self.pos[0] = float(pos_et.attrib["x"])
        self.pos[1] = float(pos_et.attrib["y"])
        self.pos[2] = float(pos_et.attrib["z"])
        
        tangent_et = path_point_et.find("./tangent")
        self.tangent[0] = float(tangent_et.attrib["x"])
        self.tangent[1] = float(tangent_et.attrib["y"])
        self.tangent[2] = float(tangent_et.attrib["z"])
        
        rotation_et = path_point_et.find("./rotation")
        self.rotation[0] = float(rotation_et.attrib["x"])
        self.rotation[1] = float(rotation_et.attrib["y"])
        self.rotation[2] = float(rotation_et.attrib["z"])
        


def indent(elem, level=0):
  i = "\n" + level*"    "
  if len(elem):
    if not elem.text or not elem.text.strip():
      elem.text = i + "    "
    if not elem.tail or not elem.tail.strip():
      elem.tail = i
    for elem in elem:
      indent(elem, level+1)
    if not elem.tail or not elem.tail.strip():
      elem.tail = i
  else:
    if level and (not elem.tail or not elem.tail.strip()):
      elem.tail = i


#def tangent(points, isCircular=False):
#    '''
#    Returns the tangent vectors of a 3-D pathline.
#
#    Parameters
#    ----------
#    points : list(numpy.array(3))
#        x-, y-, and z-coordinates of the n points defining pathline.
#    isCircular : bool
#        considers the last point of the list as coming before the first point, 
#        and the first point if the as coming after the last point.
#
#    Returns
#    -------
#    list(numpy.array(3))
#        tangent vector at each point.
#
#    '''
#    if isCircular:
#        pos = np.array(points[-1]+points+points[0])
#        d1s = np.gradient(pos, axis=0)
#        mag_d1s = np.sqrt(d1s[:,0]**2 + d1s[:,1]**2 + d1s[:,2]**2)
#        tan = np.array([1/mag_d1s] * 3).transpose() * d1s
#        
#        out = list()
#        for xi in tan[1:-1]:
#            out.append(xi)
#        
#
#    else:
#        pos = np.array(points)
#        d1s = np.gradient(pos, axis=0)
#        mag_d1s = np.sqrt(d1s[:,0]**2 + d1s[:,1]**2 + d1s[:,2]**2)
#        tan = np.array([1/mag_d1s] * 3).transpose() * d1s
#        
#        out = list()
#        for xi in tan:
#            out.append(xi)
#    
#    return out
#
#
#def RodriguesRotMatrix(axis, angle):
#    '''
#    The Rodrigues rotation formula is an efficient algorithm for rotating a 
#    vector in space, given an axis and angle of rotation. 
#    
#    Parameters
#    ----------
#    axis : numpy.array(1, 3)
#        Axis of rotation.
#    angle : float
#        Angle of rotation around the axis.
#    
#    Returns
#    -------
#    R : numpy.array(3, 3)
#        3-dimensional rotation matrix.
#    
#    References
#    ----------
#    .. https://en.wikipedia.org/wiki/Rodrigues'_rotation_formula
#
#    '''
#    a = axis/np.linalg.norm(axis)
#    a = np.nan_to_num(a)
#
#    n = np.array([[  0.0, -a[2],  a[1]],
#                  [ a[2],   0.0, -a[0]],
#                  [-a[1],  a[0],   0.0]])
#    
#    return np.eye(3) + n*np.sin(angle) + np.dot(n,n)*(1 - np.cos(angle))


def toPlaneCoordinates(path_point, points_xyz):
    '''
    Changes point coordinates from global coordinate system (base xyz) to the 
    plane coordinate system (base uvn), where n is the plane normal versor.
    
    Parameters
    ----------
    path_point : Path_Point
        A Path_Point object with the plane base vectors.
    points_xyz : list(numpy.ndarray)
        List of points in the global coodrinate system.
        
    Returns
    -------
    points_uvn : list(numpy.ndarray)
        List of points in the plane coodrinate system.
    
    See also
    --------
    toGlobalCoordinates
    '''
    
    # the norms are not exactly 1, because of truncation. I am not sure whether
    # normalising the vectors and transformation matrix improves accuracy.
    n = path_point.tangent
    n /= np.linalg.norm(n)
    
    u = path_point.rotation
    u /= np.linalg.norm(u)
    
    v = np.cross(u, n)
    v /= np.linalg.norm(v)
    
    # Transformation matrix
    T = np.stack([u,v,n], axis=1)
    T /= np.linalg.det(T)
    
    points_uvn = list()
    for point in points_xyz:
        points_uvn.append(np.linalg.solve(T, point))
    
    return points_uvn

def toGlobalCoordinates(path_point, points_uvn):
    '''
    Changes point coordinates from plane coordinate system (base uvn), where n 
    is the plane normal versor, to global coordinate system (base xyz).
    
    Parameters
    ----------
    path_point : Path_Point
        A Path_Point object with the plane base vectors.
    points_uvn : list(numpy.ndarray)
        List of points in the global coodrinate system.
        
    Returns
    -------
    points_xyz : list(numpy.ndarray)
        List of points in the plane coodrinate system.
    
    See also
    --------
    toPlaneCoordinates
    '''
    
    # the norms are not exactly 1, because of truncation. I am not sure whether
    # normalising the vectors and transformation matrix improves accuracy.
    n = path_point.tangent
    n /= np.linalg.norm(n)
    
    u = path_point.rotation
    u /= np.linalg.norm(u)
    
    v = np.cross(u, n)
    v /= np.linalg.norm(v)
    
    T = np.stack([u,v,n], axis=1)
    T /= np.linalg.det(T)
    
    points_xyz = list()
    for point in points_uvn:
        points_xyz.append(np.dot(T, point))
    
    return points_xyz

class Ellipse(object):
    """
    The Ellipse class represent an ellipse in the 2D Cartesian plane (x, y).
    """
    def __init__(self):
        #conic function parameters that defines an ellipse.
        self.a = np.zeros(6)
        self.isActive = False
    
    def fitEllipse(self, u, v):
        '''
        Fit an ellipse from n datapoints on the x-y plane using the least-squares
        method.
        
        Any conic function is defined as
    
        .. math::
            f(a, (x,y)) = D \\cdot a = 0 ,
    
        where :math:`D = (x^2, xy, y^2, x, y, 1)^T`, and 
        :math:`a = (a_{xx}, 2\,a_{xy}, a_{yy}, 2\,a_x, 2\,a_y, a_1)` is a vector 
        of function parameters that defines the conic object. The aim is to find 
        the :math:`a` vector that minimises the objective function.
    
        .. math::
            \\sum_{i=1}^n f(a, (x,y)_i)^2 
                = \\sum_{i=1}^n a^T \\, D_i^T \\, D_i \\, a 
                = a^T \\, S \\, a ,
        
        such that the vector :math:`a` is constraint for making an ellipses 
        shape, and :math:`S` is a :math:`6 \\times 6` scatter matrix. The 
        components of vector :math:`a` must satisfy
    
        .. math::
            a_{xx}a_{yy} - a_{xy}^2 = a^T \\, C \\, a > 0
    
        for the corresponding conic to be an ellipse, which happens if the 
        matrix :math:`C` is filled with zeros, except 
        :math:`C_{1,3} = C_{3,1} = 2` and :math:`C_{2,2} = -1`. Thus, the 
        constraint is used to minimise the objective function with a Lagrangian 
        multiplier :math:`\lambda`
        
        .. math::
            L(a) = a^T \\, S \\, a - \lambda \, a^T \\, C \\, a = 0 . 
        
        Hence,
        
        .. math::
            \\frac{d L}{d a} = 0 \\Rightarrow S \\, a - \\lambda \\, C \, a = 0 ,
            
        which is an eigenvalue problem
        
        .. math::
            S^{-1} \\, C a - \\frac{1}{\\lambda} \\, a = 0 ,
            
        whose parameters of the best-fit ellipse are 
        :math:`\\frac{1}{\lambda_{min}} \, a`.
        
        Parameters
        ----------
        (x, y) : (numpy.array(n), numpy.array(n))
            vectors x and y of the sampled point positions.
            
        Returns
        -------
        a : numpy.array(6)
            conic function parameters that defines an ellipse.
        
        References
        ----------
        .. http://nicky.vanforeest.com/misc/fitEllipse/fitEllipse.html
        
        .. Fitzgibbon, A.W., Pilu, M., and Fischer R.B., Direct least squares 
        fitting of ellipsees, Proc. of the 13th Internation Conference on 
        Pattern Recognition, pp 253--257, Vienna, 1996.
        
        '''
        
#        points = np.array(points)
#        x, y = np.hsplit(points, 2)
        x = u[:,np.newaxis]
        y = v[:,np.newaxis]
        
        D =  np.hstack((x*x, x*y, y*y, x, y, np.ones_like(x)))
        S = np.dot(D.T, D)
        C = np.zeros([6,6])
        C[0,2] = C[2,0] = 2.0; C[1,1] = -1.0
        
        E, V =  np.linalg.eig(np.dot(np.linalg.inv(S), C))
        n = np.argmax(np.abs(E))
        
        self.a = V[:,n]
        self.isActive = True
    
    
    def getCentre_uv(self):
        '''
        Returns the centre of the ellipse defined by the parameters of the 
        conic function.
        
        Returns
        -------
        centre : np.array(2)
            centre position in the 2D plane.
        '''
        if self.isActive:
#            a = self.a
#            axx, axy, ayy, ax, ay, a1 = a[0], a[1]/2.0, a[2], a[3]/2.0, a[4]/2.0, a[5]
            axx = self.a[0]
            axy = self.a[1]/2.0
            ayy = self.a[2]
            ax  = self.a[3]/2.0
            ay  = self.a[4]/2.0
            a1  = self.a[5]
            
            den = axx*ayy - axy*axy
        
            x0  = (axy*ay - ayy*ax)/den
            y0  = (axy*ax - axx*ay)/den
        
            return np.array([x0, y0])
        else:
            print("Ellipse hasn't been fit yet!")
    
    
    def getAxisLength(self):
        '''
        Obtains the minor and major axes of the ellipse.
        
        Returns
        -------
        axes : np.array(2)
            axes on x and y.
        
        '''
        if self.isActive:
            a = self.a
            axx, axy, ayy, ax, ay, a1 = a[0], a[1]/2.0, a[2], a[3]/2.0, a[4]/2.0, a[5]
            
            up = 2.0*(axx*ay*ay + ayy*ax*ax + a1*axy*axy - 2.0*axy*ax*ay - axx*ayy*a1)
            
            down1 =  (axy*axy - axx*ayy) \
                    *(- np.sqrt((axx - ayy)*(axx - ayy) + 4.0*axy*axy) \
                      - (axx + ayy) \
                     )
            down2 =  (axy*axy - axx*ayy) \
                    *(+ np.sqrt((axx - ayy)*(axx - ayy) + 4.0*axy*axy) \
                      - (axx + ayy) \
                     )
                    
            a = np.sqrt(np.abs(up/down1))
            b = np.sqrt(np.abs(up/down2))
            
            return np.array([a, b])
        else:
            print("Ellipse hasn't been fit yet!")
    
    
    def getAngleOfRotation(self):
        '''
        Angle or rotation of the ellipse defined as a conic function.
        
        Returns
        -------
        theta : float
            Rotation angle in radians.
        '''
        if self.isActive:
            a = self.a
            axx, axy, ayy, ax, ay, a1 = a[0], a[1]/2.0, a[2], a[3]/2.0, a[4]/2.0, a[5]
            
            if axy == 0.0:
                if axx > ayy:
                    return 0.0
                else:
                    return 0.5*np.pi
            else: 
                if axx > ayy:
                    return 0.5*np.arctan(2.0*axy/(axx - ayy))
                else:
                    return 0.5*np.pi + 0.5*np.arctan(2.0*axy/(axx - ayy))
        else:
            print("Ellipse hasn't been fit yet!")


def parse_args():
    """
    Parse command-line arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-s",
                        "--surface-file",
                        help="Input surface (.vtp) file.")
    parser.add_argument("-i",
                        "--input-contour-file",
                        help="Input contour (.ctr) file.")
    parser.add_argument("-o",
                        "--output-contour-file",
                        help="Output contour (.ctr) file.")
    parser.add_argument("-e",
                        "--edit-table-file",
                        help="Edit contour table (.csv) file.")
    args = parser.parse_args()

    if len(sys.argv) == 1:
       parser.print_help()
       sys.exit(1)

    return args


def init_logging():
    logger_name = 'sv-contours'
    logger = logging.getLogger(logger_name)
    
    # clear logger, if it is alrerady running.
    if (logger.hasHandlers()):
        logger.handlers.clear()
    
    logger.setLevel(logging.INFO)
    console_handler = logging.StreamHandler()
    formatter = logging.Formatter('[%(name)s] %(levelname)s - %(message)s')
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    
    return logger 



def read_vtp_file(file_name, logger):
    """
    Read a vtp file. 
    """

    if file_name == None:
        logger.error("No input surface file name given.")
        sys.exit(1)
    else:
        logger.info("Input surface file name: %s" % file_name)

    reader = vtk.vtkXMLPolyDataReader()
    reader.SetFileName(file_name)
    reader.Update()
    poly_data = reader.GetOutput()

    points = poly_data.GetPoints()
    num_points = points.GetNumberOfPoints()
    logger.info("Number of points %d" % num_points)

    polygons = poly_data.GetPolys()
    num_polys = polygons.GetNumberOfCells()
    logger.info("Number of polygons %d" % num_polys)

    return poly_data

def create_contour_geometry(renderer, contour_group, color=(1.0, 1.0, 1.0)):
    """ Create geometry for the contours in a contour group.
    """
    for contour in contour_group.contours:
        contour_points = contour.contour_points
        num_pts = len(contour_points)

        # Create contour geometry points and line connectivity.
        points = vtk.vtkPoints()
        points.SetNumberOfPoints(num_pts)
        lines = vtk.vtkCellArray()
        lines.InsertNextCell(num_pts + 1)
        n = 0
        for pt in contour_points:
            points.SetPoint(n, pt[0], pt[1], pt[2])
            lines.InsertCellPoint(n)
            n += 1
        #_for pt in contour_points
        lines.InsertCellPoint(0)

        geom = vtk.vtkPolyData()
        geom.SetPoints(points)
        geom.SetLines(lines)
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputData(geom)
        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        actor.GetProperty().SetColor(color)
        renderer.AddActor(actor)


def create_graphics_geometry(geom):
    """ Create geometry for display.
    """
    mapper = vtk.vtkPolyDataMapper()
    #mapper.SetInputConnection(geom)
    mapper.SetInputData(geom)
    actor = vtk.vtkActor()
    actor.SetMapper(mapper)
    #actor.GetProperty().SetLineWidth(3)
    #actor.GetProperty().EdgeVisibilityOn()
    return actor



def main():
    # Get command-line arguments.
    args = parse_args()
    contour_fileName = args.input_contour_file
    surface_fileName = args.surface_file
    
    # Set logger
    logger = init_logging()

    # Create renderer and graphics window.
    renderer = vtk.vtkRenderer()
    renderer_win = vtk.vtkRenderWindow()
    renderer_win.AddRenderer(renderer)
    renderer.SetBackground(0.5, 0.8, 0.8)
    renderer_win.SetSize(500, 500)

    # Read surface .vtp file and create render geometry.
    surface = read_vtp_file(surface_fileName, logger)
    surface_geom = create_graphics_geometry(surface)
    surface_geom.GetProperty().SetOpacity(0.5)
    renderer.AddActor(surface_geom)

    # Read the contours .ctgr file.
    contour_group = ContourGroup(logger)
    contour_group.read(contour_fileName)
    
    # Render contour group
    create_contour_geometry(renderer, contour_group)
    
    # Edit contour group
    if args.edit_table_file is not None:
        # Clone contour group
        contour_group_out = copy.deepcopy(contour_group)
        # Edit clone
        contour_group_out.edit(args.edit_table_file)
        # Render contour group
        create_contour_geometry(renderer, 
                                contour_group_out, 
                                color=(0.0, 1.0, 0.0))
        # Save contour group
        if args.output_contour_file is None:
            inputfname = args.input_contour_file.split('.')[0]
            args.output_contour_file = "./" + inputfname + "_mod.ctgr"
            logger.warning( "Output contour file name not given. " \
                           +"Using default definition: input_mod.ctgr") 
        logger.info("Output contour file name: {}".format(
                                                    args.output_contour_file))
        contour_group_out.write(args.output_contour_file)
        contour_group_out.savaData("./aorta_data.csv")
    
    #Render window
    interactor = vtk.vtkRenderWindowInteractor()
    interactor.SetInteractorStyle(vtk.vtkInteractorStyleTrackballCamera())
    interactor.SetRenderWindow(renderer_win)
    interactor.Start()
    
    del(logger)



if __name__ == "__main__":
    main()



